#ifndef AMSMF_ENCODING_H_
#define AMSMF_ENCODING_H_

#include <amsmf/Types.h>
#include <amsmf/Assert.h>
#include <amt/Defines.h>
#include <amt/Endian.h>

namespace AMSMF
{

namespace FMT
{

namespace SMF
{

typedef AMT::Endian::Big Endian;

namespace Chunk
{
typedef AMT::Endian::Field<U32> MagicF;
typedef AMT::Endian::Field<U32, MagicF> LengthF;
typedef AMT::Endian::Total<LengthF> TotalF;
}

namespace Header
{
typedef AMT::Endian::Field<U16> FormatF;
typedef AMT::Endian::Field<U16, FormatF> NumTracksF;
typedef AMT::Endian::Field<U16, NumTracksF> DivisionF;
typedef AMT::Endian::Total<DivisionF> TotalF;
const U32 MAGIC = 0x4D546864; // "MThd"
}

namespace TimeBase
{

const U16 IS_REALTIME = 0x8000U;
const U16 CLOCK_MASK = (U16)~IS_REALTIME;

inline bool
is_realtime(U16 code)
{
	return (code & IS_REALTIME) != 0;
}

inline U16
get_num_clocks(U16 code)
{
	return code & CLOCK_MASK;
}

inline U8
get_num_fps(U16 code)
{
	return (U8)(-(S8)(code >> 8) & 0x7F);
}

inline U8
get_num_subframes(U16 code)
{
	return (U8)code;
}

} // TimeBase

namespace Track
{
const U32 MAGIC = 0x4D54726B; // "MTrk"
}

namespace VarInteger
{
const int UNIT_SHIFT = 7;
const U8 CONT_MASK = 0x80;
const U8 VAL_MASK = (U8)~CONT_MASK;
const int MAX_LENGTH = 4;
}

namespace Meta
{
typedef AMT::Endian::Field<U8> KindF;
typedef AMT::Endian::Field<U8, KindF> LengthF;
typedef AMT::Endian::Total<LengthF> TotalF;
}

namespace MSG
{

// ----------------------------------------------------------------------
// MIDI
//
const U8 STATUS_MASK = 0x80;
const U8 CHMSG_KINDMASK = 0xF0;
const U8 CHMSG_CHMASK = 0x0F;

inline bool is_status_byte(U8 byte) { return (byte & STATUS_MASK) != 0; }
inline bool is_system_msg(U8 byte) { return byte >= 0xF0; }
inline bool is_ch_msg(U8 byte)
{ return is_status_byte(byte) && !is_system_msg(byte); }

inline U8 to_ch(U8 byte)
{
	AMSMF_ASSERT_U_("to_ch", is_ch_msg(byte));
	return byte & CHMSG_CHMASK;
}

inline U8 to_kind(U8 byte)
{
	AMSMF_ASSERT_U_("to_kind", is_status_byte(byte));
	return is_system_msg(byte) ? byte : byte & CHMSG_KINDMASK;
}

inline S16 decode_word_value(U8 lsb, U8 msb)
{
	return (lsb & 0x7F) | ((S16)(msb & 0x7F) << 7);
}

extern int lookup_msg_length(U8);


template <const U8 BYTE_, const int LEN_>
struct MsgBase_
{
	static const U8 STATUS = BYTE_;
	static const int LENGTH = LEN_;
	typedef AMT::Endian::Field<U8> StatusF;
	static bool check_length(const U8 *, int len)
	{ return LEN_==0 || len == LEN_; }
};

template <const U8 BYTE_, const int LEN_>
struct SysMsgBase_ : public MsgBase_<BYTE_, LEN_>
{
	static bool is_match(const U8 *ptr)
	{ return ptr[MsgBase_<BYTE_, LEN_>::StatusF::OFFSET] == BYTE_; }
};

template <const U8 BYTE_, const int LEN_>
struct ChMsgBase_ : public MsgBase_<BYTE_, LEN_>
{
	static bool is_match(const U8 *ptr)
	{ return (ptr[MsgBase_<BYTE_, LEN_>::StatusF::OFFSET] & CHMSG_KINDMASK)
	      == BYTE_;
	}
};

struct SysEx : public SysMsgBase_<0xF0, 0>
{
	typedef AMT::Endian::Field<U8, StatusF> ManufactureF;
};

struct SysMTCQuaterFrame : public SysMsgBase_<0xF1, 2>
{
	typedef AMT::Endian::Field<U8, StatusF> ValueF;
};

struct SysSongPos : public SysMsgBase_<0xF2, 3>
{
	typedef AMT::Endian::Field<U8, StatusF> PosLSBF;
	typedef AMT::Endian::Field<U8, PosLSBF> PosMSBF;
};

struct SysSongSel : public SysMsgBase_<0xF3, 2>
{
	typedef AMT::Endian::Field<U8, StatusF> NumberF;
};

struct SysTuneReq : public SysMsgBase_<0xF6, 1> { };
struct SysEoEx : public SysMsgBase_<0xF7, 0> { };
struct SysTimingClock : public SysMsgBase_<0xF8, 1> { };
struct SysStart : public SysMsgBase_<0xFA, 1> { };
struct SysContinue : public SysMsgBase_<0xFB, 1> { };
struct SysStop : public SysMsgBase_<0xFC, 1> { };
struct SysActiveSensing : public SysMsgBase_<0xFE, 1> { };
struct SysReset : public SysMsgBase_<0xFF, 1> { };

struct ChNoteOff : public ChMsgBase_<0x80, 3>
{
	typedef AMT::Endian::Field<U8, StatusF> KeyNoF;
	typedef AMT::Endian::Field<U8, KeyNoF> VelocityF;
};

struct ChNoteOn : public ChMsgBase_<0x90, 3>
{
	typedef AMT::Endian::Field<U8, StatusF> KeyNoF;
	typedef AMT::Endian::Field<U8, KeyNoF> VelocityF;
};

struct ChPolyKeyPressure : public ChMsgBase_<0xA0, 3>
{
	typedef AMT::Endian::Field<U8, StatusF> KeyNoF;
	typedef AMT::Endian::Field<U8, KeyNoF> IntensityF;
};

struct ChControlChange : public ChMsgBase_<0xB0, 3>
{
	typedef AMT::Endian::Field<U8, StatusF> NumberF;
	typedef AMT::Endian::Field<U8, NumberF> ValueF;
};

struct ChProgramChange : public ChMsgBase_<0xC0, 2>
{
	typedef AMT::Endian::Field<U8, StatusF> NumberF;
};

struct ChChannelPressure : public ChMsgBase_<0xD0, 2>
{
	typedef AMT::Endian::Field<U8, StatusF> IntensityF;
};

struct ChPitchBend : public ChMsgBase_<0xE0, 3>
{
	typedef AMT::Endian::Field<U8, StatusF> ValueLSBF;
	typedef AMT::Endian::Field<U8, ValueLSBF> ValueMSBF;
};

// ----------------------------------------------------------------------
// SMF
//

typedef SysMsgBase_<0xFF, 0> SMFMeta;

template <const U8 TYPE_, const int LEN_>
struct SMFMetaBase_ : public MsgBase_<0xFF, LEN_>
{
private:
	typedef MsgBase_<0xFF, LEN_> Super_;
public:
	static const U8 TYPE = TYPE_;
	typedef AMT::Endian::Field<U8, typename Super_::StatusF> KindF;
	typedef AMT::Endian::Field<U8, KindF> LengthF;
	typedef AMT::Endian::Total<LengthF> TotalHeaderF;
	static bool is_match(const U8 *ptr)
	{ return
	      ptr[Super_::StatusF::OFFSET] == 0xFF &&
	      ptr[KindF::OFFSET] == TYPE_;
	}
	static bool check_length(const U8 *ptr, int len)
	{
		return
		    Super_::check_length(ptr, len) &&
		    ptr[LengthF::OFFSET]+TotalHeaderF::LENGTH == len;
	}
};

struct SMFMetaSeqNo : public SMFMetaBase_<0x00, 5>
{
	typedef AMT::Endian::Field<U16, LengthF> SeqNoF;
};

struct SMFMetaText : public SMFMetaBase_<0x01, 0>
{
	typedef AMT::Endian::Field<U8, LengthF> TextF;
};

struct SMFMetaCopyright : public SMFMetaBase_<0x02, 0>
{
	typedef AMT::Endian::Field<U8, LengthF> TextF;
};

struct SMFMetaSeqTrackName : public SMFMetaBase_<0x03, 0>
{
	typedef AMT::Endian::Field<U8, LengthF> TextF;
};

struct SMFMetaInstName : public SMFMetaBase_<0x04, 0>
{
	typedef AMT::Endian::Field<U8, LengthF> TextF;
};

struct SMFMetaLyric : public SMFMetaBase_<0x05, 0>
{
	typedef AMT::Endian::Field<U8, LengthF> TextF;
};

struct SMFMetaMarker : public SMFMetaBase_<0x06, 0>
{
	typedef AMT::Endian::Field<U8, LengthF> TextF;
};

struct SMFMetaQuePoint : public SMFMetaBase_<0x07, 0>
{
	typedef AMT::Endian::Field<U8, LengthF> TextF;
};

struct SMFMetaChPrefix : public SMFMetaBase_<0x20, 4>
{
	typedef AMT::Endian::Field<U8, LengthF> ChannelF;
};

struct SMFMetaEndOfTrack : public SMFMetaBase_<0x2F, 3>
{
};

struct SMFMetaSetTempo : public SMFMetaBase_<0x51, 6>
{
	typedef AMT::Endian::Field<U8, LengthF> ValueHighF;
	typedef AMT::Endian::Field<U8, ValueHighF> ValueMidF;
	typedef AMT::Endian::Field<U8, ValueMidF> ValueLowF;
};

struct SMFMetaSMPTEOffset : public SMFMetaBase_<0x54, 8>
{
	typedef AMT::Endian::Field<U8, LengthF> HourF;
	typedef AMT::Endian::Field<U8, HourF> MinuteF;
	typedef AMT::Endian::Field<U8, MinuteF> SecondF;
	typedef AMT::Endian::Field<U8, SecondF> FrameF;
	typedef AMT::Endian::Field<U8, FrameF> SubFrameF; // 1/100 frame unit
};

struct SMFMetaTimeSign : public SMFMetaBase_<0x58, 7>
{
	typedef AMT::Endian::Field<U8, LengthF> NumeratorF;
	typedef AMT::Endian::Field<U8, NumeratorF> DenominatorF;
	typedef AMT::Endian::Field<U8, DenominatorF> NumClocksPerClickF;
	typedef AMT::Endian::Field<U8,
				   NumClocksPerClickF> Num32ndNotesPerQuater;
};

struct SMFMetaKeySign : public SMFMetaBase_<0x59, 5>
{
	typedef AMT::Endian::Field<S8, LengthF> NumSharpsF;
	typedef AMT::Endian::Field<U8, NumSharpsF> IsMinorF; // 0:major 1:minor
};

struct SMFSysEx : public SysMsgBase_<0xF0, 0>
{
	typedef AMT::Endian::Field<U8, StatusF> VarLengthF;
};

struct SMFArbitBytes : public MsgBase_<0xF7, 0>
{
	typedef AMT::Endian::Field<U8, StatusF> VarLengthF;
	static bool is_match(const U8 *) { return true; }
};


// ----------------------------------------------------------------------
// Reader
//

// MIDI
template <class M_, class Inh_>
class ReaderBase
{
protected:
	typedef M_ Msg_;
	const U8 *bytes__() const
	{ return static_cast<const Inh_ *>(this)->bytes_(); }
	template <class Field_>
	U8 fetch_() const
	{ return this->ReaderBase::bytes__()[Field_::OFFSET]; }
};

template <class Base_>
class ManufactureReaderTmpl_ : public Base_
{
public:
	U8 manufacture() const
	{ return this->template fetch_<typename Base_::Msg_::ManufactureF>(); }
};

template <class Base_>
class PosValueReaderTmpl_ : public Base_
{
public:
	S16 value() const
	{ return decode_word_value(
		this->template fetch_<typename Base_::Msg_::PosLSBF>(),
		this->template fetch_<typename Base_::Msg_::PosMSBF>()); }
};

template <class Base_>
class ChReaderTmpl_ : public Base_
{
public:
	U8 channel() const
	{ return to_ch(this->template fetch_<typename Base_::Msg_::StatusF>());
	}
};

template <class Base_>
class KeyNoReaderTmpl_ : public Base_
{
public:
	U8 key_number() const
	{ return this->template fetch_<typename Base_::Msg_::KeyNoF>(); }
};

template <class Base_>
class VelocityReaderTmpl_ : public Base_
{
public:
	U8 velocity() const
	{ return this->template fetch_<typename Base_::Msg_::VelocityF>(); }
};

template <class Base_>
class IntensityReaderTmpl_ : public Base_
{
public:
	U8 intensity() const
	{ return this->template fetch_<typename Base_::Msg_::IntensityF>(); }
};

template <class Base_>
class NumberReaderTmpl_ : public Base_
{
public:
	U8 number() const
	{ return this->template fetch_<typename Base_::Msg_::NumberF>(); }
};

template <class Base_>
class ValueReaderTmpl_ : public Base_
{
public:
	U8 value() const
	{ return this->template fetch_<typename Base_::Msg_::ValueF>(); }
};

template <class Base_>
class BendValueReaderTmpl_ : public Base_
{
public:
	S16 value() const
	{ return decode_word_value(
		this->template fetch_<typename Base_::Msg_::ValueLSBF>(),
		this->template fetch_<typename Base_::Msg_::ValueMSBF>()); }
};

template <class Msg_, class Inh_>
struct Reader
{
	typedef ReaderBase<Msg_, Inh_> Def;
};

template <class Inh_>
struct Reader<SysEx, Inh_>
{
	typedef ManufactureReaderTmpl_<ReaderBase<SysEx, Inh_> > Def;
};

template <class Inh_>
struct Reader<SysMTCQuaterFrame, Inh_>
{
	typedef ValueReaderTmpl_<ReaderBase<SysMTCQuaterFrame, Inh_> > Def;
};

template <class Inh_>
struct Reader<SysSongPos, Inh_>
{
	typedef PosValueReaderTmpl_<ReaderBase<SysSongPos, Inh_> > Def;
};

template <class Inh_>
struct Reader<SysSongSel, Inh_>
{
	typedef NumberReaderTmpl_<ReaderBase<SysSongSel, Inh_> > Def;
};

template <class Inh_>
struct Reader<ChNoteOff, Inh_>
{
	typedef ReaderBase<ChNoteOff, Inh_> A_;
	typedef ChReaderTmpl_<A_> B_;
	typedef KeyNoReaderTmpl_<B_> C_;
	typedef VelocityReaderTmpl_<C_> Def;
};

template <class Inh_>
struct Reader<ChNoteOn, Inh_>
{
	typedef ReaderBase<ChNoteOn, Inh_> A_;
	typedef ChReaderTmpl_<A_> B_;
	typedef KeyNoReaderTmpl_<B_> C_;
	typedef VelocityReaderTmpl_<C_> Def;
};

template <class Inh_>
struct Reader<ChPolyKeyPressure, Inh_>
{
	typedef ReaderBase<ChPolyKeyPressure, Inh_> A_;
	typedef ChReaderTmpl_<A_> B_;
	typedef KeyNoReaderTmpl_<B_> C_;
	typedef IntensityReaderTmpl_<C_> Def;
};

template <class Inh_>
struct Reader<ChControlChange, Inh_>
{
	typedef ReaderBase<ChControlChange, Inh_> A_;
	typedef ChReaderTmpl_<A_> B_;
	typedef NumberReaderTmpl_<B_> C_;
	typedef ValueReaderTmpl_<C_> Def;
};

template <class Inh_>
struct Reader<ChProgramChange, Inh_>
{
	typedef ReaderBase<ChProgramChange, Inh_> A_;
	typedef ChReaderTmpl_<A_> B_;
	typedef NumberReaderTmpl_<B_> Def;
};

template <class Inh_>
struct Reader<ChChannelPressure, Inh_>
{
	typedef ReaderBase<ChChannelPressure, Inh_> A_;
	typedef ChReaderTmpl_<A_> B_;
	typedef IntensityReaderTmpl_<B_> Def;
};

template <class Inh_>
struct Reader<ChPitchBend, Inh_>
{
	typedef ReaderBase<ChPitchBend, Inh_> A_;
	typedef ChReaderTmpl_<A_> B_;
	typedef BendValueReaderTmpl_<B_> Def;
};


// SMF
template <class Base_>
class MetaReaderTmpl_ : public Base_
{
public:
	U8 kind() const
	{ return this->template fetch_<typename Base_::Msg_::KindF>(); }
	U8 length() const
	{ return this->template fetch_<typename Base_::Msg_::LengthF>(); }
};

template <class Base_>
class  SeqNoReaderTmpl_ : public Base_
{
public:
	U16 number() const
	{ return SMF::Endian::From::fetch<typename Base_::Msg_::SeqNoF>(
		this->bytes__()); }
};

template <class Base_>
struct TextReaderTmpl_ : public Base_
{
	const char *text() const
	{ return (const char *)&this->bytes__()[Base_::Msg_::TextF::OFFSET]; }
};

template <class Base_>
class  MetaChReaderTmpl_ : public Base_
{
public:
	U8 channel() const
	{ return this->template fetch_<typename Base_::Msg_::ChannelF>(); }
};

template <class Base_>
class TempoReaderTmpl_ : public Base_
{
	typedef typename Base_::Msg_::ValueLowF ValueLowF_;
	typedef typename Base_::Msg_::ValueMidF ValueMidF_;
	typedef typename Base_::Msg_::ValueHighF ValueHighF_;
	template <class F_>
	U32 fetch_() const { return (U32)this->Base_::template fetch_<F_>(); }
public:
	U32 tempo() const
	{
		return
		    this->fetch_<ValueLowF_>() |
		    (this->fetch_<ValueMidF_>() << 8) |
		    (this->fetch_<ValueHighF_>() << 16);
	}
	double real_tempo() const
	{
		return 60000000.0/double(this->tempo());
	}
};

template <class Base_>
class SMPTEOffsetReaderTmpl_ : public Base_
{
public:
	U8 hour() const
	{ return this->template fetch_<typename Base_::Msg_::HourF>(); }
	U8 minute() const
	{ return this->template fetch_<typename Base_::Msg_::MinuteF>(); }
	U8 second() const
	{ return this->template fetch_<typename Base_::Msg_::SecondF>(); }
	U8 frame() const
	{ return this->template fetch_<typename Base_::Msg_::FrameF>(); }
	U8 subframe() const
	{ return this->template fetch_<typename Base_::Msg_::SubFrameF>(); }
};

template <class Base_>
class TimeSignReaderTmpl_ : public Base_
{
public:
	U8 numerator() const
	{ return this->template fetch_<typename Base_::Msg_::NumeratorF>(); }
	U8 raw_denominator() const
	{ return this->template fetch_<typename Base_::Msg_::DenominatorF>(); }
	U8 num_clocks_per_click() const
	{ return this->template fetch_<
	      typename Base_::Msg_::NumClocksPerClickF>(); }
	U8 num_32nd_notes_per_quater() const
	{ return this->template fetch_<
	      typename Base_::Msg_::Num32ndNotesPerQuater>(); }
	int calc_denominator()
	{ return 1 << this->raw_denominator(); }
};

const char *key_sign_name_(S8);
const char *major_minor_name_(U8);

template <class Base_>
class KeySignReaderTmpl_ : public Base_
{
public:
	S8 num_sharps() const
	{ return this->template fetch_<typename Base_::Msg_::NumSharpsF>(); }
	U8 is_minor() const
	{ return this->template fetch_<typename Base_::Msg_::IsMinorF>(); }
	const char *key_sign_name() const
	{ return key_sign_name_(this->num_sharps()); }
	const char *major_minor_name() const
	{ return major_minor_name_(this->is_minor()); }
};

template <class Inh_>
struct Reader<SMFMetaSeqNo, Inh_>
{
	typedef SeqNoReaderTmpl_<ReaderBase<SMFMetaSeqNo, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaText, Inh_>
{
	typedef TextReaderTmpl_<ReaderBase<SMFMetaText, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaCopyright, Inh_>
{
	typedef TextReaderTmpl_<ReaderBase<SMFMetaCopyright, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaSeqTrackName, Inh_>
{
	typedef TextReaderTmpl_<ReaderBase<SMFMetaSeqTrackName, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaInstName, Inh_>
{
	typedef TextReaderTmpl_<ReaderBase<SMFMetaInstName, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaLyric, Inh_>
{
	typedef TextReaderTmpl_<ReaderBase<SMFMetaLyric, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaMarker, Inh_>
{
	typedef TextReaderTmpl_<ReaderBase<SMFMetaMarker, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaQuePoint, Inh_>
{
	typedef TextReaderTmpl_<ReaderBase<SMFMetaQuePoint, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaChPrefix, Inh_>
{
	typedef MetaChReaderTmpl_<ReaderBase<SMFMetaChPrefix, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaSetTempo, Inh_>
{
	typedef TempoReaderTmpl_<ReaderBase<SMFMetaSetTempo, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaSMPTEOffset, Inh_>
{
	typedef SMPTEOffsetReaderTmpl_<ReaderBase<SMFMetaSMPTEOffset,
						  Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaTimeSign, Inh_>
{
	typedef TimeSignReaderTmpl_<ReaderBase<SMFMetaTimeSign, Inh_> > Def;
};

template <class Inh_>
struct Reader<SMFMetaKeySign, Inh_>
{
	typedef KeySignReaderTmpl_<ReaderBase<SMFMetaKeySign, Inh_> > Def;
};


} // MSG

} // SMF

} // FMT

} // AMSMF

#endif
