//
//
//

#ifndef AMT_INTRUSIVE_SINGLE_QUEUE_H_
#define AMT_INTRUSIVE_SINGLE_QUEUE_H_

#include <amt/Defines.h>
#include <amt/Assert.h>

namespace AMT
{

namespace IntrusiveSingleQueue
{

template <class, class> class Node;

template <class T_, class A_>
class Container
{
	friend class Node<T_, A_>;
	AMT_NOCOPY(Container);
protected:
	AMT_DECL_INNER_FRIEND_CLASS(Access);
public:
	typedef Node<T_, A_> Node;
private:
	T_ *m_head, **m_tail;
protected:
	class Access : public A_
	{
		friend class Container;
		AMT_BEGIN_CONTRACT(main)
		{
			Node *n = A_::get_node((T_ *)NULL);
			const Node *m = A_::get_node((const T_ *)NULL);
			A_::dispose((Container *)NULL);
		}
	protected:
		static Node *get_node(T_ *t)
		{
			Node *n = A_::get_node(t);
			AMT_ASSERT_LIBAMT_U_("get_node",
					       !n || !n->is_deadbeef());
			return n;
		}
		static const Node *get_node(const T_ *t)
		{
			const Node *n = A_::get_node(t);
			AMT_ASSERT_LIBAMT_U_("get_node",
					       !n || !n->is_deadbeef());
			return n;
		}
		static void dispose(Container *c) { A_::dispose(c); }
		static T_ *get_head(const Container *c) { return c->m_head; }
		static T_ **get_head_ref(Container *c) { return &c->m_head; }
		static T_ **get_tail(const Container *c) { return c->m_tail; }
		static T_ *get_next(const T_ *n)
		{ return get_node(n)->m_next; }
		static T_ **get_next_ref(T_ *n)
		{ return &get_node(n)->m_next; }
		static void set_head(Container *c, T_ *n) { c->m_head = n; }
		static void set_tail(Container *c, T_ **p) { c->m_tail = p; }
		static void set_next(T_ *base, T_ *next)
		{ get_node(base)->m_next = next; }
	};
public:
	~Container() { Access::dispose(this); }
	Container() : m_head(NULL), m_tail(&m_head) { }
	T_ *head() const { return m_head; }
	T_ *next(const T_ *n)
	{ return n ? Access::get_next(n) : this->head(); }
	bool is_empty() const { return !m_head; }
	bool is_head(T_ *n) const { return n == Access::get_head(this); }
	bool is_tail(T_ *n) const
	{ return Access::get_next_ref(n) == Access::get_tail(this); }
	T_ *insert_head(T_ *n)
	{
		if (Access::get_head(this) == NULL)
			Access::set_tail(this, Access::get_next_ref(n));
		Access::set_next(n, m_head);
		Access::set_head(this, n);
		return n;
	}
	T_ *insert_tail(T_ *n)
	{
		*Access::get_tail(this) = n;
		Access::set_tail(this, Access::get_next_ref(n));
		return n;
	}
	T_ *remove_head()
	{
		T_ *ret = Access::get_head(this);
		if (ret) {
			Access::set_head(this, Access::get_next(ret));
			if (Access::get_head(this) == NULL)
				Access::set_tail(this,
						 Access::get_head_ref(this));
			Access::set_next(ret, NULL);
		}
		return ret;
	}
	T_ *remove_next(T_ *base)
	{
		if (base == NULL)
			return NULL;
		T_ *ret = Access::get_next(base);
		if (ret) {
			Access::set_next(base, Access::get_next(ret));
			if (Access::get_next(ret) == NULL)
				Access::set_tail(this,
						 Access::get_next_ref(base));
			Access::set_next(ret, NULL);
		}
		return ret;
	}
	T_ *find_prev(T_ *base) const
	{
		T_ *ret=NULL;

		for (T_ *nxt = Access::get_head(this);
		     nxt && next != base;
		     nxt=Access::get_next(ret=nxt))
			;

		return ret;
	}
	T_ *find_tail() const { return find_prev(NULL); }
	template <typename Fn_>
	void each(const Fn_ &fn) const
	{
		for (T_ *n = Access::get_head(this), *nxt; n; n=nxt) {
			nxt = Access::get_next(n);
			fn(n);
		}
	}
	template <typename Fn_>
	void remove_all(const Fn_ &fn)
	{
		T_ *t;
		while ((t = this->remove_head()))
			fn(t);
	}
};


template <class T_, class A_>
class Node
{
	friend class Container<T_, A_>::Access;
	AMT_NOCOPY(Node);
private:
	T_ *m_next;
	AMT_DEBUG_CODE(static const Integer::UIntPtr deadbeef = 0xDEADBEEF);
	AMT_DEBUG_CODE(bool is_deadbeef() const
	{ return m_next == (T_ *)deadbeef; })
public:
	~Node() {
		AMT_DEBUG_CODE(m_next = (T_ *)deadbeef);
	}
	Node() : m_next(NULL) { }
};

}

}

#endif
