#ifndef LIBVSQ_EVENT_HANDLE_H_
#define LIBVSQ_EVENT_HANDLE_H_

#include <libvsq/Defines.h>
#include <libvsq/Types.h>
#include <libvsq/Assert.h>
#include <libvsq/Database.h>
#include <libvsq/Exception.h>
#include <amt/ScopedPtr.h>
#include <string>

namespace LibVSQ
{

namespace EventHandles
{

class VibratoCoord
{
	double m_pos;
	U8 m_value;
public:
	VibratoCoord() { }
	VibratoCoord(double p, U8 v) : m_pos(p), m_value(v) { }
	AMT_ACSOR_RW(double, pos);
	AMT_ACSOR_RW(U8, value);
};

class VibratoChanges
{
	AMT_DECL_INNER_FRIEND_CLASS(Impl_);
private:
	AMT::Ptr::Scoped<Impl_> m_impl;
	VibratoCoord &at_(int) const;
	VibratoCoord &at__(int idx) const
	{
		LIBVSQ_ASSERT_U_("operator []",
				 idx >= 0 && idx < this->num_changes());
		return this->at_(idx);
	}
public:
	VibratoChanges();
	~VibratoChanges();
	AMT_DECL_ACSOR_RW(int, num_changes);
	AMT_DECL_ACSOR_RW(U8, start_value);
	VibratoCoord &operator [] (int idx)
	{ return this->at__(idx); }
	const VibratoCoord &operator [] (int idx) const
	{ return this->at__(idx); }
};

class LyricField
{
	std::string m_key;
	std::string m_lyric;
	std::string m_phoneme;
	double m_param1;
	int m_param2;
	int m_param3;
	int m_param4;
public:
	LyricField() : m_param1(0.0), m_param2(0), m_param3(0), m_param4(0) { }
	~LyricField() { }
	AMT_ACSOR_RW(std::string, key);
	AMT_ACSOR_RW(std::string, lyric);
	AMT_ACSOR_RW(std::string, phoneme);
	AMT_ACSOR_RW(double, param1);
	AMT_ACSOR_RW(int, param2);
	AMT_ACSOR_RW(int, param3);
	AMT_ACSOR_RW(int, param4);
};

} // EventHandles

class EventHandle
{
public:
	enum Type
	{
		T_SINGER,
		T_LYRIC,
		T_VIBRATO,
	};
private:
	std::string m_name;
	Type m_type;
	bool m_is_stored;
public:
	virtual ~EventHandle() = 0;
	//
	EventHandle(const std::string &n, Type t)
		: m_name(n), m_type(t), m_is_stored(false)
	{ }
	AMT_ACSOR_R(std::string, name);
	AMT_ACSOR_RW(Type, type);
	AMT_ACSOR_RW(bool, is_stored);
	template <class Type_>
	Type_ &to()
	{
		if (this->type() != Type_::TYPE)
			throw Ex::UnmatchedType();
		return *static_cast<Type_ *>(this);
	}
	template <class Type_>
	const Type_ &to() const
	{
		if (this->type() != Type_::TYPE)
			throw Ex::UnmatchedType();
		return *static_cast<const Type_ *>(this);
	}
};
AMT_DEF_INLINE_DTOR(EventHandle);

namespace EventHandles
{

class Singer : public EventHandle
{
	std::string m_icon_id;
	std::string m_ids;
	int m_original;
	std::string m_caption;
	ClockLength m_length;
	int m_language;
	int m_program;
public:
	static const Type TYPE = T_SINGER;
	Singer(const std::string &n) : EventHandle(n, TYPE),
		   m_original(0), m_length(0), m_language(0), m_program(0)
	{ }
	~Singer() { }
	AMT_ACSOR_RW(std::string, icon_id);
	AMT_ACSOR_RW(std::string, ids);
	AMT_ACSOR_RW(int, original);
	AMT_ACSOR_RW(std::string, caption);
	AMT_ACSOR_RW(ClockLength, length);
	AMT_ACSOR_RW(int, language);
	AMT_ACSOR_RW(int, program);
};

class Lyric : public EventHandle
{
	AMT_DECL_INNER_FRIEND_CLASS(Impl_);
private:
	AMT::Ptr::Scoped<Impl_> m_impl;
	const LyricField &at_(int) const;
public:
	static const Type TYPE = T_LYRIC;
	Lyric(const std::string &);
	~Lyric();
	int num_lyrics() const;
	LyricField &insert();
	const LyricField &operator [] (int idx) const
	{
		LIBVSQ_ASSERT_U_("operator []", idx < this->num_lyrics());
		return this->at_(idx);
	}
};

class Vibrato : public EventHandle
{
	std::string m_icon_id;
	std::string m_ids;
	int m_original;
	std::string m_caption;
	ClockLength m_length;
	VibratoChanges m_depth;
	VibratoChanges m_rate;
public:
	static const Type TYPE = T_VIBRATO;
	~Vibrato() { }
	Vibrato(const std::string &n)
		: EventHandle(n, TYPE), m_original(0), m_length(0)
	{ }
	AMT_ACSOR_RW(std::string, icon_id);
	AMT_ACSOR_RW(std::string, ids);
	AMT_ACSOR_RW(int, original);
	AMT_ACSOR_RW(std::string, caption);
	AMT_ACSOR_RW(ClockLength, length);
	AMT_ACSOR_RW(VibratoChanges, depth);
	AMT_ACSOR_RW(VibratoChanges, rate);
};

class Database
{
	LibVSQ::Database<EventHandle> m_database;
public:
	Database() { }
	~Database() { }
	template <class Type_>
	Type_ *insert(const std::string &n)
	{ Type_ *d = new Type_(n); m_database.insert(n, d); return d; }
	const EventHandle &find(const std::string &n) const
	{ return m_database.find(n); }
	EventHandle &find(const std::string &n)
	{ return m_database.find(n); }
};

} // EventHandles

} // LibVSQ


#endif
