#include <libvsq/ParseVSQ.h>
#include <libvsq/ParseVSQText.h>
#include <libvsq/SortDMText.h>
#include <libvsq/Exception.h>
#include <amsmf/ParseSMF.h>
#include <string>

namespace LibVSQ
{

namespace
{
const int VSQ_SMF_FORMAT = 1;
const int VSQ_MIN_NUM_TRACKS = 2;
const int VSQ_MASTER_TRACK_NO = 0;
const char VSQ_NEWLINE = '\n';
} // <anon>

class ParseVSQ::Impl_ : public AMSMF::Parsers::Hooks, public TextParsers::Hooks
{
	ParseVSQ::Hooks &m_hooks;
	AMT::Ptr::Scoped<AMSMF::Parser> m_smf_parser;
	TextParser m_vsq_text_parser;
	SortDMText m_sorter;
	int m_num_total_tracks; // including master track
	int m_cur_track;
	ClockLength m_curpos;
	std::string m_line;
public:
	Impl_(ParseVSQ::Hooks &h)
		: m_hooks(h),
		  m_smf_parser(new AMSMF::Parser(*this)),
		  m_vsq_text_parser(*this)
	{
	}
	~Impl_() { }
	bool operator () (InputBuffer &);
	void done();
	//
	void begin_smf(const AMSMF::SMFHeader &);
	void end_smf();
	void begin_header_extra_bytes(U32) { }
	void end_header_extra_bytes() { }
	void header_extra_bytes(const U8 *, U32) { }
	void begin_track(int, U32);
	void end_track(int);
	void track_event(int, const AMSMF::Parsers::TrackEvent &);
	//
	void begin_section(const char *);
	void end_section(const char *);
	void field(const char *, const char *, const char *);
private:
	void process_master_track_(const AMSMF::Parsers::TrackEvent &);
	void process_voice_track_(const AMSMF::Parsers::TrackEvent &);
	void process_frame_text_(const char *);
};

void ParseVSQ::Impl_::begin_smf(const AMSMF::SMFHeader &hdr)
{
	if (hdr.format() != VSQ_SMF_FORMAT)
		throw Ex::IllegalSMFFormat();
	m_num_total_tracks = hdr.num_tracks();
	if (m_num_total_tracks < VSQ_MIN_NUM_TRACKS)
		throw Ex::TooFewTracks();
	if (!AMSMF::DivClock::is_match(hdr))
		throw Ex::DivisionNotSupported();
	m_hooks.begin_vsq(m_num_total_tracks-1,
			  AMSMF::DivClock(hdr).num_clocks());
}

void ParseVSQ::Impl_::end_smf()
{
	m_hooks.end_vsq();
}

void ParseVSQ::Impl_::begin_track(int tr, U32)
{
	m_curpos = 0;
	switch (tr) {
	case VSQ_MASTER_TRACK_NO:
		m_hooks.begin_master_track();
		break;
	default:
		m_cur_track = tr-1;
		std::string().swap(m_line);
		m_vsq_text_parser.reset(*this);
		m_sorter.reset();
		m_hooks.begin_voice_track(m_cur_track);
		break;
	}
}

void ParseVSQ::Impl_::end_track(int tr)
{
	switch (tr) {
	case VSQ_MASTER_TRACK_NO:
		m_hooks.end_master_track();
		break;
	default:
		if (!m_line.empty()) {
			m_vsq_text_parser.parse(&m_line[0], m_line.size());
			std::string().swap(m_line);
		}
		m_sorter.done();
		m_vsq_text_parser.done();
		m_hooks.end_voice_track(m_cur_track);
		break;
	}
}

void ParseVSQ::Impl_::process_master_track_(const TrackEvent &ev)
{
	if (AMSMF::Parsers::SMFMetaSetTempo::is_match(ev)) {
		AMSMF::Parsers::SMFMetaSetTempo rd(ev);
		m_hooks.change_tempo(m_curpos, rd.real_tempo());
	} else if (AMSMF::Parsers::SMFMetaTimeSign::is_match(ev)) {
		AMSMF::Parsers::SMFMetaTimeSign rd(ev);
		m_hooks.change_time_signature(m_curpos,
					      rd.numerator(),
					      rd.calc_denominator());
	}
}

void ParseVSQ::Impl_::process_frame_text_(const char *text)
{
	LineBuffer whole(text), line;

	while (!whole.is_empty()) {
		line = whole.skip_to(VSQ_NEWLINE);
		if (whole[0] != VSQ_NEWLINE) {
			LIBVSQ_ASSERT_I_("process_frame_text_",
					 line.is_empty());
			m_line += std::string(whole.begin(), whole.end());
			break;
		}
		if (line.is_empty())
			line = whole.cut_left(1);
		else {
			whole.cut_left(1);
			line = LineBuffer(line.begin(), line.length()+1);
		}
		if (m_line.empty()) {
			m_vsq_text_parser.parse(
				line.begin(), line.length());
		} else {
			m_line += std::string(line.begin(), line.end());
			m_vsq_text_parser.parse(&m_line[0], m_line.size());
			std::string().swap(m_line);
		}
	}
}

void
ParseVSQ::Impl_::process_voice_track_(const AMSMF::Parsers::TrackEvent &ev)
{
	if (AMSMF::Parsers::SMFMetaText::is_match(ev)) {
		AMSMF::Parsers::SMFMetaText rd(ev);
		std::string line;
		const char *text = NULL;

		try {
			text = m_sorter(rd.text());
		}
		catch (Ex::NotDMText &) {
			// just ignore.
		}
		while (text != NULL) {
			this->process_frame_text_(text);
			text = m_sorter();
		}
	}
}

void ParseVSQ::Impl_::track_event(int tr, const AMSMF::Parsers::TrackEvent &ev)
{
	m_curpos += ev.delta();
	switch (tr) {
	case VSQ_MASTER_TRACK_NO:
		this->process_master_track_(ev);
		break;
	default:
		this->process_voice_track_(ev);
		break;
	}
}

void ParseVSQ::Impl_::begin_section(const char *sect)
{
	m_hooks.begin_section(m_cur_track, sect);
}

void ParseVSQ::Impl_::end_section(const char *sect)
{
	m_hooks.end_section(m_cur_track, sect);
}

void ParseVSQ::Impl_::field(const char *sect, const char *key, const char *val)
{
	m_hooks.field(m_cur_track, sect, key, val);
}

bool ParseVSQ::Impl_::operator () (InputBuffer &ib)
{
	return m_smf_parser->parse(ib);
}

void ParseVSQ::Impl_::done()
{
	m_smf_parser->done();
}

//----------------------------------------------------------------------

ParseVSQ::ParseVSQ(Hooks &h) : m_impl(new Impl_(h)) { }

ParseVSQ::~ParseVSQ() { }

bool ParseVSQ::operator () (InputBuffer &ib)
{
	if (!m_impl.is_valid())
		return false;
	if (!ib.is_empty() && (*m_impl)(ib))
		return true;
	this->done();
	return false;
}

void ParseVSQ::done()
{
	if (m_impl.is_valid()) {
		m_impl->done();
		m_impl.reset();
	}
}

void ParseVSQ::reset(Hooks &h)
{
	this->done();
	m_impl.reset(new Impl_(h));
}

} // LibVSQ
