#include <libvsq/ReadSectHandle.h>
#include <libvsq/ReadSection.tcc>
#include <libvsq/EventDataDB.h>
#include <libvsq/EventData.h>
#include <libvsq/EventHandle.h>
#include <libvsq/LineBuffer.h>
#include <libvsq/CSV.h>
#include <amt/LexCast.h>

namespace LibVSQ
{

namespace
{

template <class Ctx_>
class Reader : public ReaderTmpl<Ctx_>
{
	typedef ReaderTmpl<Ctx_> Super_;
	typedef typename Super_::Data Data_;
public:
	Reader(Track &tr, EventHandle &h) : Super_(tr, h.to<Data_>()) { }
};

namespace Singer
{

enum Tag
{
	T_ICON_ID		= 0x00000001,
	T_IDS			= 0x00000002,
	T_ORIGINAL		= 0x00000004,
	T_CAPTION		= 0x00000008,
	T_LENGTH		= 0x00000010,
	T_LANGUAGE		= 0x00000020,
	T_PROGRAM		= 0x00000040,
};

#define SINGER_ALL(x)							      \
	x(T_ICON_ID)							      \
	x(T_IDS)							      \
	x(T_ORIGINAL)							      \
	x(T_CAPTION)							      \
	x(T_LENGTH)							      \
	x(T_LANGUAGE)							      \
	x(T_PROGRAM)

const U32 MANDATORY = LIBVSQ_RH_MAKE_MASK(SINGER_ALL);

class Context : public SimpleContextTmpl<EventHandles::Singer, Tag, MANDATORY>
{
	typedef SimpleContextTmpl<EventHandles::Singer, Tag, MANDATORY> Super_;
public:
	Context(Track &tr, Data &d) : Super_(tr, d) { }
	void check_complete() const
	{
		this->Super_::check_complete_mandatory();
		this->data().is_stored() = true;
	}
};

typedef LibVSQ::Reader<Context> Reader;

} // Singer

} // <anon>

#define RH_HANDLER(t)	HandlerTmpl<Singer::Context, Singer::t>

LIBVSQ_RH_DEF(T_ICON_ID, "IconID") { ctx.data().icon_id() = a; }

LIBVSQ_RH_DEF(T_IDS, "IDS") { ctx.data().ids() = a; }

LIBVSQ_RH_DEF(T_ORIGINAL, "Original")
{ ctx.data().original() = (int)AMT::lex_cast<long>(a); }

LIBVSQ_RH_DEF(T_CAPTION, "Caption") { ctx.data().caption() = a; }

LIBVSQ_RH_DEF(T_LENGTH, "Length")
{ ctx.data().length() = (ClockLength)AMT::lex_cast<unsigned long>(a); }

LIBVSQ_RH_DEF(T_LANGUAGE, "Language")
{ ctx.data().language() = (int)AMT::lex_cast<long>(a); }

LIBVSQ_RH_DEF(T_PROGRAM, "Program")
{ ctx.data().program() = (int)AMT::lex_cast<long>(a); }

LIBVSQ_RH_REGISTER(Singer::Context, SINGER_ALL);

#undef RH_HANDLER

namespace
{

namespace Lyric
{

class Reader : public ReadSection
{
	EventHandles::Lyric &m_lyric;
public:
	Reader(Track &, EventHandle &h) : m_lyric(h.to<EventHandles::Lyric>())
	{ }
	void operator () (const char *key, const char *val)
	{
		EventHandles::LyricField &lf = m_lyric.insert();
		CSV csv(val);

		lf.key() = key;
		lf.lyric() = csv.fetch();
		lf.phoneme() = csv.fetch();
		lf.param1() = csv.fetch_double();
		lf.param2() = csv.fetch_int();
		lf.param3() = csv.fetch_int();
		if (!csv.is_end())
			lf.param4() = csv.fetch_int();

		if (!csv.is_end())
			throw Ex::IllegalLyricField();
	}
	void check_complete() const
	{
		if (m_lyric.num_lyrics() == 0)
			throw Ex::LackedEntry();
		m_lyric.is_stored() = true;
	}
};

} // Lyric

namespace Vibrato
{

enum Tag
{
	T_ICON_ID		= 0x00000001,
	T_IDS			= 0x00000002,
	T_ORIGINAL		= 0x00000004,
	T_CAPTION		= 0x00000008,
	T_LENGTH		= 0x00000010,
	T_START_DEPTH		= 0x00000020,
	T_DEPTH_BP_NUM		= 0x00000040,
	T_DEPTH_BP_X		= 0x00000080,
	T_DEPTH_BP_Y		= 0x00000100,
	T_START_RATE		= 0x00000200,
	T_RATE_BP_NUM		= 0x00000400,
	T_RATE_BP_X		= 0x00000800,
	T_RATE_BP_Y		= 0x00001000,
};

#define VIBRATO_MANDATORY(x)						      \
	x(T_ICON_ID)							      \
	x(T_IDS)							      \
	x(T_ORIGINAL)							      \
	x(T_CAPTION)							      \
	x(T_LENGTH)							      \
	x(T_START_DEPTH)						      \
	x(T_DEPTH_BP_NUM)						      \
	x(T_START_RATE)							      \
	x(T_RATE_BP_NUM)

#define VIBRATO_DEPTH(x)						      \
	x(T_DEPTH_BP_X)							      \
	x(T_DEPTH_BP_Y)

#define VIBRATO_RATE(x)							      \
	x(T_RATE_BP_X)							      \
	x(T_RATE_BP_Y)

#define VIBRATO_ALL(x)							      \
	VIBRATO_MANDATORY(x)						      \
	VIBRATO_DEPTH(x)						      \
	VIBRATO_RATE(x)

const U32 MANDATORY = LIBVSQ_RH_MAKE_MASK(VIBRATO_MANDATORY);
const U32 DEPTH = LIBVSQ_RH_MAKE_MASK(VIBRATO_DEPTH);
const U32 RATE = LIBVSQ_RH_MAKE_MASK(VIBRATO_RATE);

class Context : public SimpleContextTmpl<EventHandles::Vibrato, Tag, MANDATORY>
{
	typedef SimpleContextTmpl<EventHandles::Vibrato, Tag, MANDATORY>
	Super_;
public:
	Context(Track &tr, Data &d) : Super_(tr, d) { }
	void check_complete() const
	{
		this->Super_::check_complete_mandatory();
		if (this->data().depth().num_changes() > 0 &&
		    (this->mask() & DEPTH) != DEPTH)
			throw Ex::LackedEntry();
		if (this->data().rate().num_changes() > 0 &&
		    (this->mask() & RATE) != RATE)
			throw Ex::LackedEntry();
		this->data().is_stored() = true;
	}
};

typedef LibVSQ::Reader<Context> Reader;

} // Vibrato

} // <anon>

#define RH_HANDLER(t)	HandlerTmpl<Vibrato::Context, Vibrato::t>

LIBVSQ_RH_DEF(T_ICON_ID, "IconID") { ctx.data().icon_id() = a; }

LIBVSQ_RH_DEF(T_IDS, "IDS") { ctx.data().ids() = a; }

LIBVSQ_RH_DEF(T_ORIGINAL, "Original")
{ ctx.data().original() = (int)AMT::lex_cast<long>(a); }

LIBVSQ_RH_DEF(T_CAPTION, "Caption") { ctx.data().caption() = a; }

LIBVSQ_RH_DEF(T_LENGTH, "Length")
{ ctx.data().length() = (ClockLength)AMT::lex_cast<unsigned long>(a); }

inline void read_start_(EventHandles::VibratoChanges &ch, const char *a)
{
	ch.start_value() = (U8)AMT::lex_cast<unsigned long>(a);
}

inline void read_num_changes_(EventHandles::VibratoChanges &ch, const char *a)
{
	ch.num_changes() = (int)AMT::lex_cast<unsigned long>(a);
}

inline void read_x_(EventHandles::VibratoChanges &ch, const char *a)
{
	CSV csv(a);
	int n = ch.num_changes();

	for (int i=0; i<n; i++)
		ch[i].pos() = csv.fetch_double();

	if (!csv.is_end())
		throw Ex::IllegalVibratoArray();
}

inline void read_y_(EventHandles::VibratoChanges &ch, const char *a)
{
	CSV csv(a);
	int n = ch.num_changes();

	for (int i=0; i<n; i++)
		ch[i].value() = (U8)csv.fetch_int();

	if (!csv.is_end())
		throw Ex::IllegalVibratoArray();
}

LIBVSQ_RH_DEF(T_START_DEPTH, "StartDepth")
{ read_start_(ctx.data().depth(), a); }

LIBVSQ_RH_DEF(T_DEPTH_BP_NUM, "DepthBPNum")
{ read_num_changes_(ctx.data().depth(), a); }

LIBVSQ_RH_DEF(T_DEPTH_BP_X, "DepthBPX")
{ read_x_(ctx.data().depth(), a); }

LIBVSQ_RH_DEF(T_DEPTH_BP_Y, "DepthBPY")
{ read_y_(ctx.data().depth(), a); }

LIBVSQ_RH_DEF(T_START_RATE, "StartRate")
{ read_start_(ctx.data().rate(), a); }

LIBVSQ_RH_DEF(T_RATE_BP_NUM, "RateBPNum")
{ read_num_changes_(ctx.data().rate(), a); }

LIBVSQ_RH_DEF(T_RATE_BP_X, "RateBPX")
{ read_x_(ctx.data().rate(), a); }

LIBVSQ_RH_DEF(T_RATE_BP_Y, "RateBPY")
{ read_y_(ctx.data().rate(), a); }

LIBVSQ_RH_REGISTER(Vibrato::Context, VIBRATO_ALL);

#undef RH_HANDLER


//

template <>
ReadSection *
ReadSectionFactoryTmpl<ReadSectHandle_>::operator () (Track &tr,
						      const char *sect) const
{
	EventHandle &h = tr.handles().find(sect);
	if (h.is_stored())
		throw Ex::DuplicatedEntry();
	switch (h.type()) {
	case EventHandle::T_SINGER:
		return new Singer::Reader(tr, h);
	case EventHandle::T_LYRIC:
		return new Lyric::Reader(tr, h);
	case EventHandle::T_VIBRATO:
		return new Vibrato::Reader(tr, h);
	}
	throw Ex::Failure();
}

template class ReadSectionFactoryTmpl<ReadSectHandle_>;

} // LibVSQ
