#include <libvsq/ReadSectMixer.h>
#include <libvsq/ReadSection.tcc>
#include <libvsq/Mixer.h>
#include <libvsq/LineBuffer.h>
#include <amt/LexCast.h>
#include <vector>

namespace LibVSQ
{

namespace
{

#define REGULAR_SET(x)							      \
	x(T_MASTER_FADER)						      \
	x(T_MASTER_PANPOT)						      \
	x(T_MASTER_MUTE)						      \
	x(T_OUTPUT_MODE)						      \
	x(T_TRACKS)

#define PER_TR_SET(x)							      \
	x(T_FADER)							      \
	x(T_PANPOT)							      \
	x(T_MUTE)							      \
	x(T_SOLO)

#define ALL_SET(x)							      \
	REGULAR_SET(x)							      \
	PER_TR_SET(x)

enum Tag
{
	T_MASTER_FADER		= 0x00000001,
	T_MASTER_PANPOT		= 0x00000002,
	T_MASTER_MUTE		= 0x00000004,
	T_OUTPUT_MODE		= 0x00000008,
	T_TRACKS		= 0x00000010,
	T_FADER			= 0x00000020,
	T_PANPOT		= 0x00000040,
	T_MUTE			= 0x00000080,
	T_SOLO			= 0x00000100,
};

const U32 T_MANDATORY = LIBVSQ_RH_MAKE_MASK(REGULAR_SET);
const U32 T_PER_TR = LIBVSQ_RH_MAKE_MASK(PER_TR_SET);

struct ArgPair
{
	int ch;
	const char *str;
};

class Context
{
public:
	typedef Mixer Data;
	typedef LibVSQ::Tag Tag;
	typedef const char *Key;
	typedef const ArgPair &Arg;
private:
	typedef std::vector<U32> PerTrMaskVec_;
	Track &m_track;
	Data &m_data;
	U32 m_mask;
	PerTrMaskVec_ m_per_tr_mask;
public:
	Context(Track &tr, Data &d) : m_track(tr), m_data(d), m_mask(0) { }
	AMT_REF_ACSOR_CONST(Track, track);
	AMT_REF_ACSOR_CONST(Data, data);
	AMT_ACSOR_R(U32, mask);
	struct CheckCompletePerTr
	{
		void operator () (U32 mask)
		{
			if ((mask & T_PER_TR) != T_PER_TR)
				throw Ex::LackedEntry();
		}
	};
	void check_complete() const
	{
		if ((m_mask & T_MANDATORY) != T_MANDATORY)
			throw Ex::LackedEntry();
		std::for_each(m_per_tr_mask.begin(), m_per_tr_mask.end(),
			      CheckCompletePerTr());
	}
	bool is_set(Tag t, Arg a) const
	{
		if (((U32)t & T_PER_TR) != 0) {
			if (a.ch < 0 || a.ch >= m_per_tr_mask.size())
				throw Ex::IllegalTrackNumber();
			return m_per_tr_mask[a.ch] & (U32)t;
		}
		return (m_mask & (U32)t) != 0;
	}
	void mark_set(Tag t, Arg a)
	{
		if (((U32)t & T_PER_TR) != 0) {
			if (a.ch < 0 || a.ch >= m_per_tr_mask.size())
				throw Ex::IllegalTrackNumber();
			m_per_tr_mask[a.ch] |= (U32)t;
		} else
			m_mask |= (U32)t;
	}
	void set_num_tracks(int ntr)
	{
		if (m_per_tr_mask.size() != 0)
			throw Ex::DuplicatedEntry();
		m_per_tr_mask.resize(ntr, 0);
	}
};

} // <anon>

#define RH_HANDLER(t)	HandlerTmpl<Context, t>

LIBVSQ_RH_DEF(T_MASTER_FADER, "MasterFeder") // what is feder?
{ ctx.data().master().fader() = (int)AMT::lex_cast<long>(a.str); }
LIBVSQ_RH_DEF(T_MASTER_PANPOT, "MasterPanpot")
{ ctx.data().master().panpot() = (int)AMT::lex_cast<long>(a.str); }
LIBVSQ_RH_DEF(T_MASTER_MUTE, "MasterMute")
{ ctx.data().master().mute() = (bool)AMT::lex_cast<long>(a.str); }
LIBVSQ_RH_DEF(T_OUTPUT_MODE, "OutputMode")
{ ctx.data().output_mode() = (int)AMT::lex_cast<long>(a.str); }
LIBVSQ_RH_DEF(T_TRACKS, "Tracks")
{
	int ntr = (int)AMT::lex_cast<long>(a.str);
	ctx.set_num_tracks(ntr);
	ctx.data().set_num_tracks(ntr);
}

LIBVSQ_RH_DEF(T_FADER, "Feder")
{ ctx.data()[a.ch].fader() = (int)AMT::lex_cast<long>(a.str); }
LIBVSQ_RH_DEF(T_PANPOT, "Panpot")
{ ctx.data()[a.ch].panpot() = (int)AMT::lex_cast<long>(a.str); }
LIBVSQ_RH_DEF(T_MUTE, "Mute")
{ ctx.data()[a.ch].mute() = (bool)AMT::lex_cast<long>(a.str); }
LIBVSQ_RH_DEF(T_SOLO, "Solo")
{ ctx.data()[a.ch].solo() = (bool)AMT::lex_cast<long>(a.str); }

LIBVSQ_RH_REGISTER(Context, ALL_SET);

class ReadSectMixer_ : public ReaderTmpl<Context>
{
	typedef ReaderTmpl<Context> Super_;
	static bool is_alpha_(char ch)
	{
		return (ch>='a' && ch<='z') || (ch>='A' && ch<='Z');
	}
	Mixer &validate_mixer_(Track &tr)
	{
		if (tr.is_mixer_valid())
			throw Ex::DuplicatedMixer();
		tr.mixer() = new Mixer;
		return *tr.mixer();
	}
public:
	ReadSectMixer_(Track &tr, const char *)
		: Super_(tr, validate_mixer_(tr)) { }
	void operator () (const char *key, const char *val)
	{
		LineBuffer keybuf(key);
		LineBuffer chbuf = keybuf.span(is_alpha_);

		ArgPair pair;
		if (chbuf.length())
			pair.ch = (int)AMT::lex_cast<long>(
				std::string(chbuf.begin(), chbuf.end()));
		else
			pair.ch = 0;
		pair.str = val;
		this->Super_::operator () (std::string(keybuf.begin(),
						       keybuf.end()).c_str(),
					   pair);
	}
};

template class ReadSectionFactoryTmpl<ReadSectMixer_>;

} // LibVSQ
