#include <libvsq/ReadTrack.h>
#include <libvsq/ReadSection.h>
#include <libvsq/Track.h>
#include <libvsq/Database.tcc>
#include <libvsq/ReadSectBPList.h>
#include <libvsq/ReadSectCommon.h>
#include <libvsq/ReadSectEventList.h>
#include <libvsq/ReadSectHandle.h>
#include <libvsq/ReadSectID.h>
#include <libvsq/ReadSectMaster.h>
#include <libvsq/ReadSectMixer.h>
#include <libvsq/LineBuffer.h>

namespace LibVSQ
{

template class Database<const ReadSectionFactory *>;

namespace
{

class Section
{
	const ReadSectionFactory &m_factory;
	bool m_mandatory;
	bool m_duplicatable;
public:
	Section(const ReadSectionFactory &f, bool m, bool d)
		: m_factory(f), m_mandatory(m), m_duplicatable(d) { }
	AMT_REF_ACSOR_CONST(const ReadSectionFactory, factory);
	AMT_ACSOR_R(bool, mandatory);
	AMT_ACSOR_R(bool, duplicatable);
};

class Sections : Database<const Section>
{
	typedef Database<const Section> Super_;
	template <class F, const bool m, const bool d>
	static Section *get_() { return new Section(F::get(), m, d); }
	template <class F>
	static Section *get_m_() { return Sections::get_<F, true, false>(); }
	template <class F>
	static Section *get_d_() { return Sections::get_<F, false, true>(); }
	template <class F>
	static Section *get_o_() { return Sections::get_<F, false, false>(); }
	Sections();
public:
	static const Sections &get() { static Sections s; return s; }
	~Sections() { }
	using Super_::num_entries;
	using Super_::operator [];
	using Super_::find_index;
	using Super_::find;
};

Sections::Sections()
{
	Database<const Section> &f = *static_cast<Super_ *>(this);
	f.insert("Common", get_m_<ReadSectCommonFactory>());
	f.insert("Master", get_o_<ReadSectMasterFactory>());
	f.insert("Mixer", get_o_<ReadSectMixerFactory>());
	f.insert("EventList", get_m_<ReadSectEventListFactory>());
	f.insert("ID", get_d_<ReadSectIDFactory>());
	f.insert("h", get_d_<ReadSectHandleFactory>());
	f.insert("PitchBendBPList", get_o_<ReadSectBPListFactory>());
	f.insert("PitchBendSensBPList", get_o_<ReadSectBPListFactory>());
	f.insert("DynamicsBPList", get_o_<ReadSectBPListFactory>());
	f.insert("EpRResidualBPList", get_o_<ReadSectBPListFactory>());
	f.insert("EpRESlopeBPList", get_o_<ReadSectBPListFactory>());
	f.insert("EpRESlopeDepthBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso1FreqBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso2FreqBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso3FreqBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso4FreqBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso1BWBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso2BWBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso3BWBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso4BWBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso1AmpBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso2AmpBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso3AmpBPList", get_o_<ReadSectBPListFactory>());
	f.insert("Reso4AmpBPList", get_o_<ReadSectBPListFactory>());
	f.insert("GenderFactorBPList", get_o_<ReadSectBPListFactory>());
	f.insert("PortamentoTimingBPList", get_o_<ReadSectBPListFactory>());
	f.insert("OpeningBPList", get_o_<ReadSectBPListFactory>());
}

} // <anon>

class ReadTrack::Impl_
{
	Track &m_track;
	ReadSection::ScopedPtr m_section;
	int m_curidx;
	std::vector<bool> m_set;
public:
	Impl_(Track &tr) : m_track(tr), m_curidx(-1) { }
	~Impl_() { }
	void begin_section(const char *);
	void end_section();
	void field(const char *, const char *);
	void check_complete() const;
};

void
ReadTrack::Impl_::begin_section(const char *sect)
{
	if (m_section.is_valid())
		throw Ex::IllegalCallSequence();

	const Sections &ss = Sections::get();
	LineBuffer pfx = LineBuffer(sect).skip_to('#');

	if (pfx.is_empty())
		pfx = sect;

	m_curidx = ss.find_index(std::string(pfx.begin(), pfx.end()));
	const Section &s = ss[m_curidx];
	if (m_set.size() == 0)
		m_set.resize(ss.num_entries(), false);
	LIBVSQ_ASSERT_I_("begin_section", m_set.size() <= m_curidx);
	if (!s.duplicatable() && m_set[m_curidx])
		throw Ex::DuplicatedEntry();
	m_section.reset(s.factory()(m_track, sect));
}

void
ReadTrack::Impl_::end_section()
{
	if (!m_section.is_valid())
		throw Ex::IllegalCallSequence();
	m_section.reset();
	m_set[m_curidx] = true;
	m_curidx = -1;
}

void
ReadTrack::Impl_::field(const char *key, const char *val)
{
	if (!m_section.is_valid())
		throw Ex::IllegalCallSequence();
	(*m_section)(key, val);
}

void
ReadTrack::Impl_::check_complete() const
{
	const Sections &ss = Sections::get();
	int n = ss.num_entries();

	if (m_section.is_valid())
		throw Ex::IllegalCallSequence();
	if (m_set.size() < n)
		throw Ex::LackedEntry();

	for (int i=0; i<n; i++)
		if (ss[i].mandatory() && !m_set[i])
			throw Ex::LackedEntry();
}

ReadTrack::ReadTrack(Track &tr) : m_impl(new Impl_(tr)) { }

ReadTrack::~ReadTrack() { }

void ReadTrack::begin_section(const char *sect)
{ m_impl->begin_section(sect); }

void ReadTrack::end_section() { m_impl->end_section(); }

void ReadTrack::field(const char *key, const char *val)
{ m_impl->field(key, val); }

void ReadTrack::check_complete() const { m_impl->check_complete(); }

} // LibVSQ
