#include <libvsq/Song.h>
#include <libvsq/Track.h>
#include <libvsq/EventDataDB.h>
#include <libvsq/EventList.h>
#include <libvsq/EventData.h>
#include <libvsq/EventHandle.h>
#include <libvsq/Mixer.h>
#include <libvsq/BPList.h>
#include <libvsq/TempoMap.h>
#include <libvsq/TimeSignMap.h>
#include <libvsq/ReadSong.h>
#include <amsmf/InputStream.h>
#include <stdio.h>
#include <stdarg.h>

static int indent = 0;

void
print(const char *fmt, ...)
{
	va_list ap;
	printf("%*s", indent*2, "");
	va_start(ap, fmt);
	vfprintf(stdout, fmt, ap);
	va_end(ap);
}

void
print(const LibVSQ::EventHandles::Singer &sh)
{
	print("icon id: %s\n", sh.icon_id().c_str());
	print("ids: %s\n", sh.ids().c_str());
	print("original: %d\n", sh.original());
	print("caption: %s\n", sh.caption().c_str());
	print("length: %u\n", (unsigned)sh.length());
	print("lang: %d\n", sh.language());
	print("prog: %d\n", sh.program());
}

void
print(const LibVSQ::EventHandles::Lyric &lh)
{
	int nl = lh.num_lyrics();
	for (int i=0; i<nl; i++) {
		const LibVSQ::EventHandles::LyricField &lf = lh[i];
		print("%s:\n", lf.key().c_str());
		indent++;
		print("lyrics: %s\n", lf.lyric().c_str());
		print("phoneme: %s\n", lf.phoneme().c_str());
		print("param1: %f\n", lf.param1());
		print("param2: %d\n", lf.param2());
		print("param3: %d\n", lf.param3());
		print("param4: %d\n", lf.param4());
		indent--;
	}
}

void
print(const LibVSQ::EventHandles::VibratoChanges &ch)
{
	print("start: %d\n", (int)ch.start_value());
	int n = (int)ch.num_changes();
	print("num changes: %d\n", n);
	indent++;
	for (int i=0; i<n; i++) {
		const LibVSQ::EventHandles::VibratoCoord &co = ch[i];
		print("(%f, %u)\n", co.pos(), co.value());
	}
	indent--;
}

void
print(const LibVSQ::EventHandles::Vibrato &vh)
{
	print("icon id: %s\n", vh.icon_id().c_str());
	print("ids: %s\n", vh.ids().c_str());
	print("original: %d\n", vh.original());
	print("caption: %s\n", vh.caption().c_str());
	print("length: %u\n", (unsigned)vh.length());
	print("depth:\n");
	indent++;
	print(vh.depth());
	indent--;
	print("rate:\n");
	indent++;
	print(vh.rate());
	indent--;
}

void
print(const LibVSQ::EventDatas::Singer &s)
{
	print("Type: Singer\n");
	print("IconHandle: %s\n", s.icon_handle().name().c_str());
	indent++;
	print(s.icon_handle().to<LibVSQ::EventHandles::Singer>());
	indent--;
}

void
print(const LibVSQ::EventDatas::ANote &n)
{
	print("Type: Anote\n");
	print("Length: %u\n", (unsigned)n.length());
	print("Note#: %u\n", (unsigned)n.note_no());
	print("Dynamics: %u\n", (unsigned)n.dynamics());
	print("PMBendDepth: %u\n", (unsigned)n.pm_bend_depth());
	print("PMBendLength: %u\n", (unsigned)n.pm_bend_length());
	print("PMbPortamentoUse: %u\n", (unsigned)n.portament_use());
	print("DEMdecGainRate: %u\n", (unsigned)n.decay_rate());
	print("DEMaccent: %u\n", (unsigned)n.accent());
	print("LyricHandle: %s\n", n.lyric_handle().name().c_str());
	indent++;
	print(n.lyric_handle().to<LibVSQ::EventHandles::Lyric>());
	indent--;
	if (n.is_vibrato_valid()) {
		print("VibratoHandle: %s\n",n.vibrato_handle().name().c_str());
		indent++;
		print(n.vibrato_handle().to<LibVSQ::EventHandles::Vibrato>());
		indent--;
		print("VibratoDelay: %u\n", (unsigned)n.vibrato_delay());
	} else
		print("no Vibrato\n");
}

void
print(const LibVSQ::EventDatas::EndOfSong &)
{
	print("End Of Song\n");
}

void
print(const LibVSQ::EventData &ed)
{
	switch (ed.type()) {
	case LibVSQ::EventData::T_SINGER:
		print(ed.to<LibVSQ::EventDatas::Singer>());
		break;
	case LibVSQ::EventData::T_ANOTE:
		print(ed.to<LibVSQ::EventDatas::ANote>());
		break;
	case LibVSQ::EventData::T_EOS:
		print(ed.to<LibVSQ::EventDatas::EndOfSong>());
		break;
	}
}

void
print(const LibVSQ::EventList &l)
{
	print("[EventList]\n");
	indent++;
	int nev = l.num_entries();
	for (int i=0; i<nev; i++) {
		print("%d: %s\n",
		      (int)l[i].pos(), l[i].entry()->name().c_str());
		indent++;
		print(l[i].entry()->data());
		indent--;
	}
	indent--;
}

void
print(const LibVSQ::Tracks::Common &c)
{
	print("[common]\n");
	indent++;
	print("ver: %s\n", c.version().c_str());
	print("name: %s\n", c.name().c_str());
	print("color: %d,%d,%d\n", c.color().r, c.color().g, c.color().b);
	print("dm: %d\n", c.dynamics_mode());
	print("pm: %d\n", c.play_mode());
	indent--;
}

void
print(const LibVSQ::Tracks::Master &m)
{
	print("[Master]\n");
	indent++;
	print("premeasure: %d\n", (int)m.pre_measure());
	indent--;
}

void
print(const LibVSQ::Mixer &m)
{
	print("[Mixer]\n");
	indent++;
	print("master fader   %d\n", m.master().fader());
	print("master panpot: %d\n", m.master().panpot());
	print("master mute:   %d\n", (int)m.master().mute());
	print("output mode:   %d\n", (int)m.output_mode());
	int ntr = m.num_tracks();
	for (int i=0; i<ntr; i++) {
		print("tr: %d\n", i);
		indent++;
		const LibVSQ::Mixer::Param &p = m[i];
		print("fader:  %d\n", p.fader());
		print("panpot: %d\n", p.panpot());
		print("mute:   %d\n", (int)p.mute());
		print("solo:   %d\n", (int)p.solo());
		indent--;
	}
	indent--;
}

void
print(const LibVSQ::BPList &bpl)
{
	print("[%s]\n", bpl.name().c_str());
	indent++;
	int n = bpl.num_entries();
	for (int i=0; i<n; i++) {
		const LibVSQ::BPList::Pair &p = bpl[i];
		print("%u=%d\n", (unsigned)p.pos(), p.entry());
	}
	indent--;
}

void
print(const LibVSQ::BPLists::Database &d)
{
	int n = d.num_entries();
	for (int i=0; i<n; i++)
		print(d[i]);
}

void
print(const LibVSQ::Track &tr)
{
	print(tr.common());
	print(tr.events());
	print(tr.bplists());
}

void
print(const LibVSQ::TempoMap &tm)
{
	print("TempoMap:\n");
	indent++;
	int n = tm.num_entries();
	for (int i=0; i<n; i++) {
		const LibVSQ::TempoMap::Pair &t = tm[i];
		print("%u=%f\n", (unsigned)t.pos(), t.entry().tempo());
	}
	indent--;
}

void
print(const LibVSQ::TimeSignMap &tm)
{
	print("TimeSignMap:\n");
	indent++;
	int n = tm.num_entries();
	for (int i=0; i<n; i++) {
		const LibVSQ::TimeSignMap::Pair &t = tm[i];
		print("%u=%d/%d\n", (unsigned)t.pos(),
		      t.entry().numerator(), t.entry().denominator());
	}
	indent--;
}

void
print(const LibVSQ::Song &s)
{
	print("division=%u\n", (unsigned)s.division());
	if (s.is_master_valid())
		print(s.master());
	if (s.is_mixer_valid())
		print(s.mixer());
	print(s.tempo_map());
	print(s.time_sign_map());
	int ntr=s.num_tracks();
	for (int i=0; i<ntr; i++) {
		print("track %d:\n", i+1);
		indent++;
		print(s.track(i));
		indent--;
	}
}

int
main(int argc, char **argv)
{
	if (argc != 2) {
		fprintf(stderr, "usage: test_ParseSMF000 <file>\n");
		return 1;
	}
	AMSMF::InputStream is(argv[1]);
	LibVSQ::ReadSong rs;

	while (rs(is.read()))
		;

	AMT::Ptr::Scoped<const LibVSQ::Song> song(rs.done());
	print(*song);

	return 0;
}
