#include <amsmf/ParseSMF.h>
#include <amsmf/InputBuffer.h>
#include <amsmf/InputStream.h>
#include <libvsq/SortDMText.h>
#include <libvsq/Exception.h>
#include <libvsq/ParseVSQ.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

using namespace AMT;
using namespace AMSMF;

namespace Ex
{

using namespace LibVSQ::Ex;
using namespace AMSMF::Ex;

struct NoText : public DataError { };

}

class VSQHooks : public LibVSQ::ParseVSQ::Hooks
{
	FILE *m_out;
	bool m_out_need_close;
	int m_num_voice_tracks;
	LibVSQ::ClockLength m_division;
	bool m_is_there_dm_text;
	static void chkprf_(int ret)
	{
		if (ret < 0)
			throw ::Ex::CannotWriteFile();
	}
public:
	VSQHooks() : m_out(stdout), m_out_need_close(false) { }
	~VSQHooks()
	{
		if (m_out_need_close)
			fclose(m_out);
	}
	void begin_vsq(int ntr, LibVSQ::ClockLength div)
	{
		m_num_voice_tracks = ntr;
		m_division = div;
	}
	void end_vsq()
	{
	}
	void begin_master_track()
	{
		chkprf_(fprintf(m_out, "[Master]\n"));
		chkprf_(fprintf(m_out, "NumVoices=%d\n", m_num_voice_tracks));
		chkprf_(fprintf(m_out, "Division=%d\n", (int)m_division));
		chkprf_(fprintf(m_out, "[TempoMap]\n"));
	}
	void end_master_track()
	{
		chkprf_(fprintf(m_out, "\n"));
	}
	void change_tempo(LibVSQ::ClockLength pos, double tmp)
	{
		chkprf_(fprintf(m_out, "%u=Tempo:%.2f\n", (unsigned)pos, tmp));
	}
	void change_time_signature(LibVSQ::ClockLength pos, int num, int den)
	{
		chkprf_(fprintf(m_out, "%u=Time:%u/%u:24:8\n",
				(unsigned)pos, (unsigned)num, (unsigned)den));
	}
	void begin_voice_track(int tr)
	{
		m_is_there_dm_text = false;
		chkprf_(fprintf(m_out, "[Track%02d]\n", tr+1));
	}
	void end_voice_track(int tr)
	{
		if (!m_is_there_dm_text)
			throw ::Ex::NoText();
		if (tr != m_num_voice_tracks-1)
			chkprf_(fprintf(m_out, "\n"));
	}
	void begin_section(int, const char *sect)
	{
		m_is_there_dm_text = true;
		chkprf_(fprintf(m_out, "[%s]\n", sect));
	}
	void end_section(int, const char *)
	{
	}
	void field(int, const char *, const char *key, const char *val)
	{
		chkprf_(fprintf(m_out, "%s=%s\n", key, val));
	}
};

void
usage()
{
	fprintf(stderr, "usage: vsqdump <file>\n");
	exit(EXIT_FAILURE);
}

void
err(const char *msg)
{
	fprintf(stderr, "Error: %s", msg);
	exit(EXIT_FAILURE);
}

int
main(int argc, char **argv)
{
	if (argc != 2)
		usage();

	try {
		InputStream s(argv[1]);
		VSQHooks h;
		LibVSQ::ParseVSQ pvsq(h);
		while (pvsq(s.read()))
			;
		if (!s.read().is_empty())
			err("There are needless trailing bytes.\n");
	}
	catch (::Ex::IllegalSMFFormat &) {
		err("Illegal SMF Format (Format 1 required).\n");
	}
	catch (::Ex::TooFewTracks &) {
		err("Too few number of tracks.\n");
	}
	catch (::Ex::DivisionNotSupported &) {
		err("SMPTE style division is not supported.\n");
	}
	catch (LibVSQ::Ex::LackedDMFrame &) {
		err("Some DM frames lacked.\n");
	}
	catch (LibVSQ::Ex::DuplicatedDMFrame &) {
		err("Some DM frames lacked.\n");
	}
	catch (::Ex::NoText &) {
		err("No DM frames - SMF, but not VSQ.\n");
	}
	catch (::Ex::UnexpectedFileEnd &) {
		err("Unexpected file end.\n");
	}
	catch (::Ex::IllegalMagic &) {
		err("Illegal SMF magic.\n");
	}
	catch (::Ex::ChunkOverrun &) {
		err("SMF data error - track chunk overrun.\n");
	}
	catch (::Ex::IllegalRunningStatus &) {
		err("SMF data error - illegal running status.\n");
	}
	catch (::Ex::TooShortHeader &) {
		err("Too short SMF header.\n");
	}
	catch (::Ex::IllegalSystemMessage &) {
		err("SMF data error - raw system message is disallowed.\n");
	}
	catch (::Ex::IllegalStatusByte &) {
		err("SMF data error - illegal status byte.\n");
	}
	catch (::Ex::TooLongVarInteger &) {
		err("SMF data error - variable length integer overflow.\n");
	}
	catch (::Ex::CannotOpenFile &) {
		err("cannot open file.\n");
	}
	catch (::Ex::CannotReadFile &) {
		err("cannot read file.\n");
	}
	catch (::Ex::CannotWriteFile &) {
		err("cannot write file.\n");
	}
	catch (...) {
		err("exception caught.\n");
	}

	return 0;
}
